"""Selects time slice"""
import logging
from typing import Union

import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("select_id")
class SelectIDIterDataPipe(IterDataPipe):
    """Selects time slice"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        location_datapipe: IterDataPipe,
        data_source_name: str = "nwp",
    ):
        """
        Selects id

        Args:
            source_datapipe: Datapipe of Xarray objects
            location_datapipe: Location datapipe
            data_source_name: the name of the datasource.
        """
        self.source_datapipe = source_datapipe
        self.location_datapipe = location_datapipe
        self.data_source_name = data_source_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, location in self.source_datapipe.zip_ocf(self.location_datapipe):

            logger.debug(f"Selecting Data on id {location.id} for {self.data_source_name}")

            if self.data_source_name == "nwp":
                try:
                    xr_data = xr_data.sel(id=location.id)
                except Exception as e:
                    logger.warning(f"Could not find {location.id} in nwp {xr_data.id}")
                    raise e

            if self.data_source_name == "pv":
                try:
                    xr_data = xr_data.sel(pv_system_id=[location.id])
                except Exception as e:
                    logger.warning(f"Could not find {location.id} in pv {xr_data.pv_system_id}")
                    raise e
            logger.debug("Selected Data on id")
            yield xr_data



"""Various utilites that didn't fit elsewhere"""
import logging
from pathlib import Path
from typing import Iterator, Optional, Sequence, Sized, Tuple, Union

import fsspec.asyn
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from pandas.core.dtypes.common import is_datetime64_dtype
from pathy import Pathy
from torch.utils.data.datapipes._decorator import functional_datapipe
from torch.utils.data.datapipes.datapipe import IterDataPipe
from torch.utils.data.datapipes.iter.combining import T_co

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch

logger = logging.getLogger(__name__)


def datetime64_to_float(datetimes: np.ndarray, dtype=np.float64) -> np.ndarray:
    """
    Converts datetime64 to floats

    Args:
        datetimes: Array of datetimes
        dtype: Dtype to convert to

    Returns:
        Converted datetimes to floats
    """
    nums = datetimes.astype("datetime64[s]").astype(dtype)
    mask = np.isfinite(datetimes)
    return np.where(mask, nums, np.NaN)


def assert_num_dims(tensor, num_expected_dims: int) -> None:
    """
    Asserts the tensor shape is correct

    Args:
        tensor: Tensor to check
        num_expected_dims: Number of expected dims
    """
    assert len(tensor.shape) == num_expected_dims, (
        f"Expected tensor to have {num_expected_dims} dims." f" Instead, shape={tensor.shape}"
    )


def is_sorted(array: np.ndarray) -> bool:
    """Return True if array is sorted in ascending order."""
    # Adapted from https://stackoverflow.com/a/47004507/732596
    if len(array) == 0:
        return False
    return np.all(array[:-1] <= array[1:])


def check_path_exists(path: Union[str, Path]):
    """Raise a FileNotFoundError if `path` does not exist.

    `path` can include wildcards.
    """
    if not path:
        raise FileNotFoundError("Not a valid path!")
    filesystem = get_filesystem(path)
    if not filesystem.exists(path):
        # Maybe `path` includes a wildcard? So let's use `glob` to check.
        # Try `exists` before `glob` because `glob` might be slower.
        files = filesystem.glob(path)
        if len(files) == 0:
            raise FileNotFoundError(f"{path} does not exist!")


def get_filesystem(path: Union[str, Path]) -> fsspec.AbstractFileSystem:
    r"""Get the fsspect FileSystem from a path.

    For example, if `path` starts with `gs://` then return a gcsfs.GCSFileSystem.

    It is safe for `path` to include wildcards in the final filename.
    """
    path = Pathy(path)
    return fsspec.open(path.parent).fs


def sample_row_and_drop_row_from_df(
    df: pd.DataFrame, rng: np.random.Generator
) -> tuple[pd.Series, pd.DataFrame]:
    """Return sampled_row, dataframe_with_row_dropped."""
    assert not df.empty
    row_idx = rng.integers(low=0, high=len(df))
    row = df.iloc[row_idx]
    df = df.drop(row.name)
    return row, df


def set_fsspec_for_multiprocess() -> None:
    """
    Clear reference to the loop and thread.

    This is a nasty hack that was suggested but NOT recommended by the lead fsspec developer!

    This appears necessary otherwise gcsfs hangs when used after forking multiple worker processes.
    Only required for fsspec >= 0.9.0

    See:
    - https://github.com/fsspec/gcsfs/issues/379#issuecomment-839929801
    - https://github.com/fsspec/filesystem_spec/pull/963#issuecomment-1131709948

    TODO: Try deleting this two lines to make sure this is still relevant.
    """
    fsspec.asyn.iothread[0] = None
    fsspec.asyn.loop[0] = None
    fsspec.asyn._lock = None


def stack_np_examples_into_batch(np_examples: Sequence[NumpyBatch]) -> NumpyBatch:
    """
    Stacks Numpy examples into a batch

    Args:
        np_examples: Numpy examples to stack

    Returns:
        The stacked NumpyBatch object
    """
    np_batch: NumpyBatch = {}
    batch_keys = np_examples[0]  # Batch keys should be the same across all examples.
    for batch_key in batch_keys:
        if batch_key.name.endswith("t0_idx") or batch_key == BatchKey.nwp_channel_names:
            # These are always the same for all examples.
            np_batch[batch_key] = np_examples[0][batch_key]
        else:
            examples_for_key = [np_example[batch_key] for np_example in np_examples]
            try:
                np_batch[batch_key] = np.stack(examples_for_key)
            except Exception as e:
                logger.debug(f"Could not stack the following shapes together, ({batch_key})")
                shapes = [example_for_key.shape for example_for_key in examples_for_key]
                logger.debug(shapes)
                logger.error(e)
                raise e
    return np_batch


def select_time_periods(
    xr_data: Union[xr.DataArray, xr.Dataset], time_periods: pd.DataFrame, dim_name: str = "time_utc"
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Selects time periods from Xarray object

    Args:
        xr_data: Xarray object
        time_periods: Time periods to select
        dim_name: Dimension name for time

    Returns:
        The subselected Xarray object
    """
    new_xr_data = []
    for _, row in time_periods.iterrows():
        start_dt = row["start_dt"]
        end_dt = row["end_dt"]
        new_xr_data.append(xr_data.sel({dim_name: slice(start_dt, end_dt)}))
    return xr.concat(new_xr_data, dim=dim_name)


def pandas_periods_to_our_periods_dt(
    periods: Union[Sequence[pd.Period], pd.PeriodIndex]
) -> pd.DataFrame:
    """
    Converts Pandas periods to new periods

    Args:
        periods: Pandas periods to convert

    Returns:
        Converted pandas periods
    """
    new_periods = []
    for period in periods:
        new_periods.append(dict(start_dt=period.start_time, end_dt=period.end_time))
    return pd.DataFrame(new_periods)


# https://github.com/pytorch/data/issues/865
@functional_datapipe("zip_ocf")
class ZipperIterDataPipe(IterDataPipe[Tuple[T_co]]):
    """
    Aggregates elements into a tuple from each of the input DataPipes (functional name: ``zip``).

    The output is stopped as soon as the shortest input DataPipe is exhausted.

    Args:
        *datapipes: Iterable DataPipes being aggregated
    Example:
        >>> # xdoctest: +REQUIRES(module:torchdata)
        >>> from torchdata.datapipes.iter import IterableWrapper
        >>> dp1, dp2, dp3 = IterableWrapper(range(5)), IterableWrapper(range(10, 15)),
        >>> IterableWrapper(range(20, 25))
        >>> list(dp1.zip(dp2, dp3))
        [(0, 10, 20), (1, 11, 21), (2, 12, 22), (3, 13, 23), (4, 14, 24)]
    """

    datapipes: Tuple[IterDataPipe]
    length: Optional[int]

    def __init__(self, *datapipes: IterDataPipe):
        """Init"""
        if not all(isinstance(dp, IterDataPipe) for dp in datapipes):
            raise TypeError(
                "All inputs are required to be `IterDataPipe` " "for `ZipIterDataPipe`."
            )
        super().__init__()
        self.datapipes = datapipes  # type: ignore[assignment]
        self.length = None

    def __iter__(self) -> Iterator[Tuple[T_co]]:
        """Iter"""
        iterators = [iter(datapipe) for datapipe in self.datapipes]
        for data in zip(*iterators):
            yield data

    def __len__(self) -> int:
        """Len"""
        if self.length is not None:
            if self.length == -1:
                raise TypeError("{} instance doesn't have valid length".format(type(self).__name__))
            return self.length
        if all(isinstance(dp, Sized) for dp in self.datapipes):
            self.length = min(len(dp) for dp in self.datapipes)
        else:
            self.length = -1
        return len(self)


def _trig_transform(values: np.ndarray, period: Union[float, int]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Given a list of values and an upper limit on the values, compute trig decomposition.

    Args:
        values: ndarray of points in the range [0, period]
        period: period of the data

    Returns:
        Decomposition of values into sine and cosine of data with given period
    """

    return np.sin(values * 2 * np.pi / period), np.cos(values * 2 * np.pi / period)


def trigonometric_datetime_transformation(datetimes: npt.ArrayLike) -> np.ndarray:
    """
    Given an iterable of datetimes, returns a trigonometric decomposition on hour, day and month

    Args:
        datetimes: ArrayLike of datetime64 values

    Returns:
        Trigonometric decomposition of datetime into hourly, daily and
        monthly values.
    """
    assert is_datetime64_dtype(datetimes), "Data for Trig Decomposition must be np.datetime64 type"

    datetimes = pd.DatetimeIndex(datetimes)
    hour = datetimes.hour.values.reshape(-1, 1) + (datetimes.minute.values.reshape(-1, 1) / 60)
    day = datetimes.day.values.reshape(-1, 1)
    month = datetimes.month.values.reshape(-1, 1)

    sine_hour, cosine_hour = _trig_transform(hour, 24)
    sine_day, cosine_day = _trig_transform(day, 366)
    sine_month, cosine_month = _trig_transform(month, 12)

    return np.concatenate(
        [sine_month, cosine_month, sine_day, cosine_day, sine_hour, cosine_hour], axis=1
    )



"""Data pipeline for loading PV, NWP and GSP"""
import logging
from datetime import timedelta
from pathlib import Path
from typing import Union

import xarray
from torchdata.datapipes.iter import IterDataPipe

import ocf_datapipes  # noqa
from ocf_datapipes.batch import MergeNumpyModalities
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenConfiguration, OpenNWP, OpenPVFromNetCDF, OpenSatellite
from ocf_datapipes.utils.consts import NWP_MEAN, NWP_STD

logger = logging.getLogger(__name__)
xarray.set_options(keep_attrs=True)

# should scale with batch_size #TODO
BUFFER_SIZE = 100


def pv_nwp_satellite_data_pipeline(configuration: Union[Path, str, Configuration]) -> IterDataPipe:
    """
    Make data pipe with PV, NWP and Satellite

    The location can be made either from GSP or PV

    Args:
        configuration: the configuration filename for the pipe, can also be the actual configuration

    Returns: datapipe
    """

    # load configuration
    if type(configuration) != Configuration:
        config_datapipe = OpenConfiguration(configuration)
        configuration: Configuration = next(iter(config_datapipe))

    # Load NWP data
    logger.debug("Load NWP data")
    nwp_datapipe = OpenNWP(configuration.input_data.nwp.nwp_zarr_path)

    # Load PV data
    logger.debug("Load PV data")
    pv_datapipe, pv_location_datapipe = (
        OpenPVFromNetCDF(pv=configuration.input_data.pv).pv_fill_night_nans().fork(2)
    )

    logger.debug("Load Satellite data")
    satellite_datapipe = OpenSatellite(
        zarr_path=configuration.input_data.satellite.satellite_zarr_path
    )

    # add satellite data
    logger.debug("Add t0 idx")
    pv_datapipe = pv_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
    )

    logger.debug("Add t0 idx")
    nwp_datapipe = nwp_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(hours=1),
        history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
    )

    logger.debug("Add t0 idx")
    satellite_datapipe = satellite_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.satellite.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
    )

    # Pick locations
    location_datapipes = pv_location_datapipe.location_picker().fork(4, buffer_size=BUFFER_SIZE)

    # take PV space slice
    pv_datapipe, pv_time_periods_datapipe, pv_t0_datapipe = pv_datapipe.select_spatial_slice_meters(
        location_datapipe=location_datapipes[1],
        roi_height_meters=configuration.input_data.pv.pv_image_size_meters_height,
        roi_width_meters=configuration.input_data.pv.pv_image_size_meters_width,
        y_dim_name="y_osgb",
        x_dim_name="x_osgb",
    ).fork(3)
    # take NWP space slice
    nwp_datapipe, nwp_time_periods_datapipe = nwp_datapipe.select_spatial_slice_pixels(
        location_datapipe=location_datapipes[2],
        roi_height_pixels=configuration.input_data.nwp.nwp_image_size_pixels_height,
        roi_width_pixels=configuration.input_data.nwp.nwp_image_size_pixels_width,
        y_dim_name="y_osgb",
        x_dim_name="x_osgb",
    ).fork(2)

    # take Satellite space slice
    (
        satellite_datapipe,
        satellite_time_periods_datapipe,
    ) = satellite_datapipe.select_spatial_slice_pixels(
        location_datapipe=location_datapipes[3],
        roi_height_pixels=configuration.input_data.satellite.satellite_image_size_pixels_height,
        roi_width_pixels=configuration.input_data.satellite.satellite_image_size_pixels_width,
        y_dim_name="y_geostationary",
        x_dim_name="x_geostationary",
    ).fork(
        2
    )

    # get time periods
    # get contiguous time periods
    logger.debug("Getting contiguous time periods")
    nwp_time_periods_datapipe = nwp_time_periods_datapipe.get_contiguous_time_periods(
        sample_period_duration=timedelta(hours=3),  # Init times are 3 hours apart
        history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
        time_dim="init_time_utc",
    )
    pv_time_periods_datapipe = pv_time_periods_datapipe.get_contiguous_time_periods(
        sample_period_duration=timedelta(minutes=5),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
    )
    satellite_time_periods_datapipe = satellite_time_periods_datapipe.get_contiguous_time_periods(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.satellite.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.satellite.forecast_minutes),
    )

    # find joint overlapping time periods
    logger.debug("Getting joint time periods")
    overlapping_datapipe = pv_time_periods_datapipe.select_overlapping_time_slice(
        secondary_datapipes=[
            nwp_time_periods_datapipe,
            satellite_time_periods_datapipe,
        ],
    )
    gsp_time_periods, nwp_time_periods, pv_time_periods = overlapping_datapipe.fork(
        3, buffer_size=BUFFER_SIZE
    )
    # select time periods
    pv_t0_datapipe = pv_t0_datapipe.select_time_periods(time_periods=gsp_time_periods)

    # select t0 periods
    logger.debug("Select t0 joint")
    (
        nwp_t0_datapipe,
        pv_t0_datapipe,
        satellite_t0_datapipe,
    ) = pv_t0_datapipe.select_t0_time().fork(3)

    # take nwp time slices
    logger.debug("Take NWP time slices")
    nwp_datapipe = (
        nwp_datapipe.convert_to_nwp_target_time(
            t0_datapipe=nwp_t0_datapipe,
            sample_period_duration=timedelta(hours=1),
            history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
            forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
        )
        .normalize(mean=NWP_MEAN, std=NWP_STD)
        .convert_nwp_to_numpy_batch()
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )

    # take pv time slices
    logger.debug("Take PV time slices")
    pv_datapipe = (
        pv_datapipe.select_time_slice(
            t0_datapipe=pv_t0_datapipe,
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
            forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
        )
        .convert_pv_to_numpy_batch()
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )

    # take pv time slices
    logger.debug("Take Satellite time slices")
    satellite_datapipe = (
        satellite_datapipe.select_time_slice(
            t0_datapipe=satellite_t0_datapipe,
            sample_period_duration=timedelta(
                minutes=configuration.input_data.satellite.time_resolution_minutes
            ),
            history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
            forecast_duration=timedelta(
                minutes=configuration.input_data.satellite.forecast_minutes
            ),
        )
        .convert_satellite_to_numpy_batch()
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )

    ####################################
    # Join data pipes together, and get extra details
    #####################################
    logger.debug("Combine all the data sources")
    combined_datapipe = (
        MergeNumpyModalities([nwp_datapipe, pv_datapipe, satellite_datapipe])
        # .encode_space_time()
        # .add_sun_position(modality_name="gsp")
    )

    return combined_datapipe



"""Create the training/validation datapipe for training the national MetNet/-2 Model"""
import datetime
import logging
from pathlib import Path
from typing import Union

import xarray
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.convert import ConvertGSPToNumpy
from ocf_datapipes.select import DropGSP, LocationPicker
from ocf_datapipes.training.common import (
    add_selected_time_slices_from_datapipes,
    get_and_return_overlapping_time_periods_and_t0,
    open_and_return_datapipes,
)
from ocf_datapipes.transform.xarray import PreProcessMetNet
from ocf_datapipes.utils.consts import NWP_MEAN, NWP_STD, SAT_MEAN, SAT_MEAN_DA, SAT_STD, SAT_STD_DA

xarray.set_options(keep_attrs=True)
logger = logging.getLogger("metnet_datapipe")
logger.setLevel(logging.DEBUG)


def normalize_gsp(x):  # So it can be pickled
    """
    Normalize the GSP data

    Args:
        x: Input DataArray

    Returns:
        Normalized DataArray
    """
    return x / x.capacity_megawatt_power


def normalize_pv(x):  # So it can be pickled
    """
    Normalize the GSP data

    Args:
        x: Input DataArray

    Returns:
        Normalized DataArray
    """
    return x / x.capacity_watt_power


def _remove_nans(x):
    return x.fillna(0.0)


def metnet_national_datapipe(
    configuration_filename: Union[Path, str],
    use_sun: bool = True,
    use_nwp: bool = True,
    use_sat: bool = True,
    use_hrv: bool = True,
    use_pv: bool = False,
    use_gsp: bool = True,
    use_topo: bool = True,
    output_size: int = 256,
    gsp_in_image: bool = False,
    start_time: datetime.datetime = datetime.datetime(2014, 1, 1),
    end_time: datetime.datetime = datetime.datetime(2023, 1, 1),
) -> IterDataPipe:
    """
    Make GSP national data pipe

    Currently only has GSP and NWP's in them

    Args:
        configuration_filename: the configruation filename for the pipe
        use_sun: Whether to add sun features or not
        use_pv: Whether to use PV input or not
        use_hrv: Whether to use HRV Satellite or not
        use_sat: Whether to use non-HRV Satellite or not
        use_nwp: Whether to use NWP or not
        use_topo: Whether to use topographic map or not
        use_gsp: Whether to use GSP history
        start_time: Start time to select on
        end_time: End time to select from
        output_size: Size, in pixels, of the output image
        gsp_in_image: Add GSP history as channels in MetNet image

    Returns: datapipe
    """

    # load datasets
    used_datapipes = open_and_return_datapipes(
        configuration_filename=configuration_filename,
        use_nwp=use_nwp,
        use_topo=use_topo,
        use_sat=use_sat,
        use_hrv=use_hrv,
        use_gsp=use_gsp,
        use_pv=use_pv,
    )
    # Load GSP national data
    used_datapipes["gsp"] = used_datapipes["gsp"].select_train_test_time(start_time, end_time)

    # Now get overlapping time periods
    used_datapipes = get_and_return_overlapping_time_periods_and_t0(used_datapipes)

    # And now get time slices
    used_datapipes = add_selected_time_slices_from_datapipes(used_datapipes)

    # Now do the extra processing
    gsp_history = used_datapipes["gsp"].normalize(normalize_fn=normalize_gsp)
    gsp_datapipe = used_datapipes["gsp_future"].normalize(normalize_fn=normalize_gsp)
    # Split into GSP for target, only national, and one for history
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])

    if "nwp" in used_datapipes.keys():
        # take nwp time slices
        logger.debug("Take NWP time slices")
        nwp_datapipe = used_datapipes["nwp"].normalize(mean=NWP_MEAN, std=NWP_STD)

    if "sat" in used_datapipes.keys():
        logger.debug("Take Satellite time slices")
        # take sat time slices
        sat_datapipe = used_datapipes["sat"].normalize(mean=SAT_MEAN_DA, std=SAT_STD_DA)

    if "hrv" in used_datapipes.keys():
        logger.debug("Take HRV Satellite time slices")
        sat_hrv_datapipe = used_datapipes["hrv"].normalize(mean=SAT_MEAN["HRV"], std=SAT_STD["HRV"])

    if "topo" in used_datapipes.keys():
        topo_datapipe = used_datapipes["topo"].map(_remove_nans)

    # Now combine in the MetNet format
    modalities = []
    if gsp_in_image and "hrv" in used_datapipes.keys():
        sat_hrv_datapipe, sat_gsp_datapipe = sat_hrv_datapipe.fork(2)
        gsp_history = gsp_history.drop_gsp(gsps_to_keep=[0]).create_gsp_image(
            image_datapipe=sat_gsp_datapipe
        )
    elif gsp_in_image and "sat" in used_datapipes.keys():
        sat_datapipe, sat_gsp_datapipe = sat_datapipe.fork(2)
        gsp_history = gsp_history.drop_gsp(gsps_to_keep=[0]).create_gsp_image(
            image_datapipe=sat_gsp_datapipe
        )
    if "nwp" in used_datapipes.keys():
        modalities.append(nwp_datapipe)
    if "hrv" in used_datapipes.keys():
        modalities.append(sat_hrv_datapipe)
    if "sat" in used_datapipes.keys():
        modalities.append(sat_datapipe)
    if "topo" in used_datapipes.keys():
        modalities.append(topo_datapipe)
    if gsp_in_image:
        modalities.append(gsp_history)

    gsp_datapipe, gsp_loc_datapipe = gsp_datapipe.fork(2, buffer_size=5)

    location_datapipe = LocationPicker(gsp_loc_datapipe)

    metnet_datapipe = PreProcessMetNet(
        modalities,
        location_datapipe=location_datapipe,
        center_width=500_000,
        center_height=1_000_000,
        context_height=10_000_000,
        context_width=10_000_000,
        output_width_pixels=output_size,
        output_height_pixels=output_size,
        add_sun_features=use_sun,
    )
    gsp_datapipe = ConvertGSPToNumpy(gsp_datapipe)

    if not gsp_in_image:
        gsp_history = gsp_history.map(_remove_nans)
        gsp_history = ConvertGSPToNumpy(gsp_history, return_id=True)
        return metnet_datapipe.zip_ocf(gsp_history, gsp_datapipe)  # Makes (Inputs, Label) tuples
    else:
        return metnet_datapipe.zip(gsp_datapipe)



from datetime import timedelta

import numpy as np

from ocf_datapipes.transform.xarray import AddT0IdxAndSamplePeriodDuration, PVPowerRemoveZeroData


def test_pv_power_remove_data(passiv_datapipe):
    passiv_datapipe = AddT0IdxAndSamplePeriodDuration(
        passiv_datapipe,
        history_duration=timedelta(minutes=60),
        sample_period_duration=timedelta(minutes=5),
    )
    data_before = next(iter(passiv_datapipe))

    print(data_before)

    # set some valyes to zeros
    # data_before.values[0] = 0

    passiv_datapipe = PVPowerRemoveZeroData(passiv_datapipe, window=timedelta(minutes=60))
    data = next(iter(passiv_datapipe))

    assert np.isnan(data.values[0, 0])
    assert len(data.time_utc) == len(data_before.time_utc)



""" Make fake NWP data """
from datetime import datetime

import numpy as np

from ocf_datapipes.batch.fake.utils import get_n_time_steps_from_config, make_time_utc
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.utils.consts import BatchKey


def make_fake_nwp_data(configuration: Configuration, t0_datetime_utc: datetime):
    """
    Make Fake NWP data ready for ML model

    Args:
        configuration: configuration object
        t0_datetime_utc: one datetime for when t0 is

    Returns: dictionary of nwp items
    """

    nwp_config = configuration.input_data.nwp
    if nwp_config is None:
        return {}

    batch_size = configuration.process.batch_size
    n_channels = len(nwp_config.nwp_channels)
    n_y_osgb = nwp_config.nwp_image_size_pixels_height
    n_x_osgb = nwp_config.nwp_image_size_pixels_width
    n_fourier_features = 8

    # make time matrix
    time_utc = make_time_utc(
        batch_size=batch_size,
        history_minutes=nwp_config.history_minutes,
        forecast_minutes=nwp_config.forecast_minutes,
        t0_datetime_utc=t0_datetime_utc,
        time_resolution_minutes=nwp_config.time_resolution_minutes,
    )
    n_times = time_utc.shape[1]

    # main nwp components
    batch = {}
    batch[BatchKey.nwp_init_time_utc] = time_utc  # Seconds since UNIX epoch (1970-01-01).
    batch[BatchKey.nwp_target_time_utc] = time_utc  # Seconds since UNIX epoch (1970-01-01).
    batch[BatchKey.nwp] = np.random.random((batch_size, n_times, n_channels, n_y_osgb, n_x_osgb))
    batch[BatchKey.nwp_t0_idx] = get_n_time_steps_from_config(
        input_data_configuration=nwp_config, include_forecast=False
    )

    batch[BatchKey.nwp_step] = np.random.randint(0, 100, (batch_size, n_times))
    batch[BatchKey.nwp_y_osgb] = np.random.randint(0, 100, (batch_size, n_y_osgb))
    batch[BatchKey.nwp_x_osgb] = np.random.randint(0, 100, (batch_size, n_x_osgb))
    batch[BatchKey.nwp_channel_names] = np.random.randint(0, 100, (n_channels,))

    # fourier components
    batch[BatchKey.nwp_x_osgb_fourier] = np.random.random(
        (batch_size, n_x_osgb, n_fourier_features)
    )
    batch[BatchKey.nwp_y_osgb_fourier] = np.random.random(
        (batch_size, n_y_osgb, n_fourier_features)
    )
    batch[BatchKey.nwp_target_time_utc] = np.random.random(
        (batch_size, n_times, n_fourier_features)
    )
    batch[BatchKey.nwp_init_time_utc] = np.random.random((batch_size, n_times, n_fourier_features))

    return batch



from datetime import timedelta

from ocf_datapipes.transform.xarray import AddT0IdxAndSamplePeriodDuration, PVPowerRollingWindow


def test_pv_power_rolling_window_passiv(passiv_datapipe):
    passiv_datapipe = AddT0IdxAndSamplePeriodDuration(
        passiv_datapipe,
        history_duration=timedelta(minutes=60),
        sample_period_duration=timedelta(minutes=5),
    )
    data_before = next(iter(passiv_datapipe))
    passiv_datapipe = PVPowerRollingWindow(passiv_datapipe, expect_dataset=False)
    data = next(iter(passiv_datapipe))
    assert len(data.time_utc) == len(data_before.time_utc)



"""Convert point PV sites to image output"""
import logging
from typing import Union

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils import Zipper

logger = logging.getLogger(__name__)


@functional_datapipe("create_gsp_image")
class CreateGSPImageIterDataPipe(IterDataPipe):
    """Create GSP image from individual sites"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        image_datapipe: IterDataPipe,
        normalize: bool = False,
        image_dim: str = "geostationary",
        always_return_first: bool = False,
        seed=None,
    ):
        """
        Creates a 3D data cube of GSP output image x number of timesteps

        This is primarily for national GSP, so single GSP inputs are preferred

        Args:
            source_datapipe: Source datapipe of PV data
            image_datapipe: Datapipe emitting images to get the shape from, with coordinates
            normalize: Whether to normalize based off the image max, or leave raw data
            image_dim: Dimension name for the x and y dimensions
            always_return_first: Always return the first image data cube, to save computation
                Only use for if making the image at the beginning of the stack
            seed: Random seed to use if using max_num_pv_systems
        """
        self.source_datapipe = source_datapipe
        self.image_datapipe = image_datapipe
        self.normalize = normalize
        self.x_dim = "x_" + image_dim
        self.y_dim = "y_" + image_dim
        self.rng = np.random.default_rng(seed=seed)
        self.always_return_first = always_return_first

    def __iter__(self) -> xr.DataArray:
        for gsp_systems_xr, image_xr in Zipper(self.source_datapipe, self.image_datapipe):
            # Create empty image to use for the PV Systems, assumes image has x and y coordinates
            pv_image = np.zeros(
                (
                    len(gsp_systems_xr["time_utc"]),
                    len(image_xr[self.y_dim]),
                    len(image_xr[self.x_dim]),
                ),
                dtype=np.float32,
            )
            for i, gsp_system_id in enumerate(gsp_systems_xr["gsp_id"]):
                gsp_system = gsp_systems_xr.sel(gsp_id=gsp_system_id)
                for time_step in range(len(gsp_system.time_utc.values)):
                    # Now go by the timestep to create cube of GSP data
                    pv_image[time_step:, :] = gsp_system.isel(time_utc=time_step).values

            pv_image = np.nan_to_num(pv_image)

            # Should return Xarray as in Xarray transforms
            # Same coordinates as the image xarray, so can take that
            pv_image = _create_data_array_from_image(pv_image, gsp_systems_xr, image_xr)
            yield pv_image


def _create_data_array_from_image(
    pv_image: np.ndarray,
    pv_systems_xr: Union[xr.Dataset, xr.DataArray],
    image_xr: Union[xr.Dataset, xr.DataArray],
):
    data_array = xr.DataArray(
        data=pv_image,
        coords=(
            ("time_utc", pv_systems_xr.time_utc.values),
            ("y_geostationary", image_xr.y_geostationary.values),
            ("x_geostationary", image_xr.x_geostationary.values),
        ),
        name="gsp_image",
    ).astype(np.float32)
    data_array.attrs = image_xr.attrs
    return data_array



"""Converts Satellite to int8 for Power Perceiver"""
import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("convert_satellite_to_int8")
class ConvertSatelliteToInt8IterDataPipe(IterDataPipe):
    """Converts satellite to int8"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Converts satellite to int8

        Args:
            source_datapipe: Datapipe of satellite data
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> xr.DataArray:
        for xr_dataset in self.source_datapipe:
            xr_dataset = xr_dataset.clip(min=0, max=1023)
            xr_dataset.data = (xr_dataset.astype(np.float32).data / 4.0).round().astype(np.uint8)
            yield xr_dataset



"""Convert point PV sites to image output"""
import logging
from typing import Optional

import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("remove_northern_gsp")
class RemoveNorthernGSPIterDataPipe(IterDataPipe):
    """Remove northern GSPs"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        northern_y_osgb_limit: Optional[int] = 1_036_975,
    ):
        """
        Remove northern GSP. This might be because satellite data is not available

        Args:
            source_datapipe: Source datapipe of GSP data
            northern_y_osgb_limit: limit all gsp above this limit.
                The deafult gets rid of the very northern ones
        """
        self.source_datapipe = source_datapipe
        self.northern_y_osgb_limit = northern_y_osgb_limit

    def __iter__(self) -> xr.DataArray:
        for source_datapipe in self.source_datapipe:

            logger.debug(
                f"Removing any gsp with Y OSGB greater than {self.northern_y_osgb_limit}. "
                f"There are currently {len(source_datapipe.gsp_id)} GSPs"
            )

            keep_index = source_datapipe.y_osgb < self.northern_y_osgb_limit
            keep_gsp_ids = source_datapipe.gsp_id[keep_index]

            source_datapipe = source_datapipe.sel(gsp_id=keep_gsp_ids)
            logger.debug(f"There are now {len(source_datapipe.gsp_id)} GSPs")

            yield source_datapipe



"""Select spatial slices"""
import logging
from typing import Union

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import Location
from ocf_datapipes.utils.geospatial import load_geostationary_area_definition_and_transform_osgb

logger = logging.getLogger(__name__)


@functional_datapipe("select_spatial_slice_pixels")
class SelectSpatialSlicePixelsIterDataPipe(IterDataPipe):
    """Select spatial slice based off pixels from point of interest"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        location_datapipe: IterDataPipe,
        roi_height_pixels: int,
        roi_width_pixels: int,
        y_dim_name: str = "y",
        x_dim_name: str = "x",
    ):
        """
        Select spatial slice based off pixels from point of interest

        Args:
            source_datapipe: Datapipe of Xarray data
            location_datapipe: Location datapipe
            roi_height_pixels: ROI height in pixels
            roi_width_pixels: ROI width in pixels
            y_dim_name: Dimension name for Y
            x_dim_name: Dimension name for X
        """
        self.source_datapipe = source_datapipe
        self.location_datapipe = location_datapipe
        self.roi_height_pixels = roi_height_pixels
        self.roi_width_pixels = roi_width_pixels
        self.y_dim_name = y_dim_name
        self.x_dim_name = x_dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, location in self.source_datapipe.zip_ocf(self.location_datapipe):
            if "geostationary" in self.x_dim_name:
                center_idx: Location = _get_idx_of_pixel_closest_to_poi_geostationary(
                    xr_data=xr_data,
                    center_osgb=location,
                    x_dim_name=self.x_dim_name,
                    y_dim_name=self.y_dim_name,
                )
            else:
                center_idx: Location = _get_idx_of_pixel_closest_to_poi(
                    xr_data=xr_data,
                    center_osgb=location,
                    x_dim_name=self.x_dim_name,
                    y_dim_name=self.y_dim_name,
                )

            # Compute the index for left and right:
            half_height = self.roi_height_pixels // 2
            half_width = self.roi_width_pixels // 2

            left_idx = int(center_idx.x - half_width)
            right_idx = int(center_idx.x + half_width)
            top_idx = int(center_idx.y - half_height)
            bottom_idx = int(center_idx.y + half_height)

            # Sanity check!
            assert left_idx >= 0, f"{left_idx=} must be >= 0!"
            data_width_pixels = len(xr_data[self.x_dim_name])
            assert right_idx <= data_width_pixels, f"{right_idx=} must be <= {data_width_pixels=}"
            assert top_idx >= 0, f"{top_idx=} must be >= 0!"
            data_height_pixels = len(xr_data[self.y_dim_name])
            assert (
                bottom_idx <= data_height_pixels
            ), f"{bottom_idx=} must be <= {data_height_pixels=}"

            selected = xr_data.isel(
                {
                    self.x_dim_name: slice(left_idx, right_idx),
                    self.y_dim_name: slice(top_idx, bottom_idx),
                }
            )
            yield selected


@functional_datapipe("select_spatial_slice_meters")
class SelectSpatialSliceMetersIterDataPipe(IterDataPipe):
    """Select spatial slice based off meters from point of interest

    Currently assumes that there is pv_system_id to use isel on
    """

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        location_datapipe: IterDataPipe,
        roi_height_meters: int,
        roi_width_meters: int,
        dim_name: str = "pv_system_id",
        y_dim_name: str = "y_osgb",
        x_dim_name: str = "x_osgb",
    ):
        """
        Select spatial slice based off pixels from point of interest

        Args:
            source_datapipe: Datapipe of Xarray data
            location_datapipe: Location datapipe
            roi_height_meters: ROI height in meters
            roi_width_meters: ROI width in meters
            dim_name: Dimension name to select for ID
            y_dim_name: the y dimension name, this is so we can switch between osgb and lat,lon
            x_dim_name: the x dimension name, this is so we can switch between osgb and lat,lon
        """
        self.source_datapipe = source_datapipe
        self.location_datapipe = location_datapipe
        self.roi_height_meters = roi_height_meters
        self.roi_width_meters = roi_width_meters
        self.dim_name = dim_name
        self.y_dim_name = y_dim_name
        self.x_dim_name = x_dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, location in self.source_datapipe.zip_ocf(self.location_datapipe):
            # Compute the index for left and right:
            logger.debug("Getting Spatial Slice Meters")

            half_height = self.roi_height_meters // 2
            half_width = self.roi_width_meters // 2

            left = location.x - half_width
            right = location.x + half_width
            bottom = location.y - half_height
            top = location.y + half_height
            # Select data in the region of interest:
            id_mask = (
                (left <= getattr(xr_data, self.x_dim_name))
                & (getattr(xr_data, self.x_dim_name) <= right)
                & (getattr(xr_data, self.y_dim_name) <= top)
                & (bottom <= getattr(xr_data, self.y_dim_name))
            )

            selected = xr_data.isel({self.dim_name: id_mask})
            yield selected


def _get_idx_of_pixel_closest_to_poi(
    xr_data: xr.DataArray, center_osgb: Location, y_dim_name: str = "y", x_dim_name: str = "x"
) -> Location:
    """
    Return x and y index location of pixel at center of region of interest.

    Args:
        xr_data: Xarray dataset
        center_osgb: Center in OSGB coordinates
        y_dim_name: Y dimension name
        x_dim_name: X dimension name

    Returns:
        The Location for the center pixel
    """
    y_index = xr_data.get_index(y_dim_name)
    x_index = xr_data.get_index(x_dim_name)
    return Location(
        y=y_index.get_indexer([float(center_osgb.y)], method="nearest")[0],
        x=x_index.get_indexer([float(center_osgb.x)], method="nearest")[0],
    )


def _get_idx_of_pixel_closest_to_poi_geostationary(
    xr_data: xr.DataArray,
    center_osgb: Location,
    x_dim_name="x_geostationary",
    y_dim_name="y_geostationary",
) -> Location:
    """
    Return x and y index location of pixel at center of region of interest.

    Args:
        xr_data: Xarray dataset
        center_osgb: Center in OSGB coordinates
        x_dim_name: X dimension name
        y_dim_name: Y dimension name

    Returns:
        Location for the center pixel in geostationary coordinates
    """
    _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
    center_geostationary_tuple = _osgb_to_geostationary(xx=center_osgb.x, yy=center_osgb.y)
    center_geostationary = Location(
        x=center_geostationary_tuple[0], y=center_geostationary_tuple[1]
    )

    # Get the index into x and y nearest to x_center_geostationary and y_center_geostationary:
    x_index_at_center = np.searchsorted(xr_data[x_dim_name].values, center_geostationary.x) - 1
    # y_geostationary is in descending order:
    y_index_at_center = len(xr_data[y_dim_name]) - (
        np.searchsorted(xr_data[y_dim_name].values[::-1], center_geostationary.y) - 1
    )
    return Location(x=x_index_at_center, y=y_index_at_center)
